@@ -19,6 +19,8 @@ try: import multiprocessing
 except ImportError: pass
 import getopt
 import sys
+try: import sysconfig
+except ImportError: pass
 import os
 import platform
 import glob
@@ -49,7 +51,7 @@ def exit_on_Usage(func, *args, **kargs):
     try:
         return func(*args, **kargs)
     except Usage, err:
-        print_debug(''.join([PROGRAM_NAME, ':']), str(err.msg))
+        print_debug('{0!s}: {1!s}'.format(PROGRAM_NAME, err.msg))
         if locals().has_key('options'):
             if options.has_key('h') or options.has_key('help'): return 2
         print_debug("for help use --help")
@@ -85,6 +87,7 @@ class Environment(object):
         self._options = self.parse_argv()
         self._set_verbosity()
         self._exit_if_needed()
+        
         self._set_initial_num_cpus()
         self._check_if_dry_run()
         return
@@ -106,6 +109,8 @@ class Environment(object):
 
     def set_target_dir(self, target_dir):
         self.target_dir = target_dir
+        if self.is_debug():
+            print_debug('Setting Environment target_dir to {0!s}'.format(target_dir))
         return
         
     def get_source_dir(self):
@@ -113,6 +118,8 @@ class Environment(object):
 
     def set_source_dir(self, source_dir):
         self.source_dir = source_dir
+        if self.is_debug():
+            print_debug('Setting Environment source_dir to {0!s}'.format(source_dir))
         return
 
     @exit_on_Usage
@@ -137,7 +144,11 @@ class Environment(object):
         leaves_in_source_dir = leaves(source_dir)
         for leaf in leaves_in_source_dir:
             if self._is_valid_file(leaf):
-                parsed_filenames.append(filename_parser(leaf))
+                try:
+                    appendable = filename_parser(leaf)
+                    parsed_filenames.append(appendable)
+                except InvalidFileException:
+                    pass
         self._sequence.extend(parsed_filenames)
         return parsed_filenames
 
@@ -147,12 +158,15 @@ class Environment(object):
         '''
         options = self.get_options()
         # Always update the sequence before getting it
+        find = self._options.has_key('find')
         if self._is_first_time:
-            if self._options.has_key('find'):
-                self.find_files()
+            if find: self.find_files()
             self._update_sequence()
             self._is_first_time = False
         sequence = self._sequence
+        if len(sequence) == 0:
+            if not find: # find anyway
+                self.find_files()
         return sequence
 
     def _update_sequence(self, additional_sequence=[]):
@@ -177,7 +191,7 @@ class Environment(object):
             try:
                 self._sequence.append(filename_parser(f))
             except InvalidFileException:
-                continue
+                pass
         return
             
     def _check_if_dry_run(self):
@@ -412,14 +426,14 @@ class Environment(object):
         
         if with_args:
             if self.is_debug():
-                print_debug('Using', str(self._filename_parser.__name__),
-                            'with kwargs',
-                            str(self.get_fp_kwargs()))
+                print_debug('Using {0!s} with kwargs {1!s}'.format(
+                                self._filename_parser.__name__,
+                                self.get_fp_kwargs()))
             return partial(self._filename_parser, **self.get_fp_kwargs())
         else:
             if self.is_debug():
-                print_debug('Using', str(self._filename_parser.__name__),
-                            'without kwargs')
+                print_debug('Using {0!s} without kwargs'.format(
+                                            self._filename_parser.__name__))
             return self._filename_parser
 
     def set_num_cpus(self, num):
@@ -512,7 +526,7 @@ class Environment(object):
                 return False
     
     @exit_on_Usage 
-    def do_action(self, action):
+    def do_action(self, action, stay_open=False):
         '''
         executes an action
         
@@ -520,7 +534,7 @@ class Environment(object):
         '''
         sequence = list(self.get_sequence())
         if len(sequence) == 0:
-            raise Usage('No input files specified. Nothing to do.')
+            raise Usage('No input files specified or found. Nothing to do.')
         if not self._allow_action:
             print_debug('Test run. Nothing done.')
             print_debug('I would have acted on these files:',
@@ -534,6 +548,10 @@ class Environment(object):
     
         used_cpus = min([self.get_num_cpus(), max_cpus])
     
+        # Create output directory if it doesn't exist
+        for item in sequence:
+            item.check_output_dir(item.output_dir)
+        
         if used_cpus == 1:
             if self.is_debug(): print_debug('multiprocessing disabled')
             for item in sequence:
@@ -552,48 +570,47 @@ class Environment(object):
             results = [p.apply_async(action, (item,), kwargs) for
                        item in sequence]
             stdouts = (result.get() for result in results)
-            stdouts_good = itertools.ifilter(lambda x: x is not None, stdouts)
-            if not self.is_silent() and stdouts is not None:
+            stdouts_good = filter(lambda x: type(x) is str, stdouts)
+            if not self.is_silent() and stdouts_good is not None:
                 print >>sys.stdout, os.linesep.join(stdouts_good)
-                
-        return self.execute_next_script()
+
+        if not stay_open:
+            sys.exit(0)
+        elif self.next_script is not None:
+            return self.execute_next_script()
     
     def execute_next_script(self):
         '''
         execute the next script
-        
-        if there isn't one, then die
         '''
-        if self.next_script is not None:
-            if debug: print_debug('Launching the next script', NEXT_SCRIPT)
-            # always proceed with --find
-            next_script_args = ["--find"]
-            # pass along verbosity level
-            for verbosity in self.verbosity_levels:
-                exec(''.join(["next_script_args.append('", verbosity, "')"]))
-            os.execlp(self.next_script, "--find")
-        else:
-            sys.exit(0)
+        if debug: print_debug('Launching the next script', self.next_script)
+        # always proceed with --find
+        next_script_args = ["--find"]
+        # pass along verbosity level
+        for verbosity in self.verbosity_levels:
+            exec(''.join(["next_script_args.append('", verbosity, "')"]))
+        os.execlp(self.next_script, "--find")
     
 def _quote(s):
     return ''.join(["'", s ,"'"])
 
 class Usage(Exception):
     def __init__(self, *args):
-        self.msg = ''.join(args)
+        self.msg = ''.join(map(str(args)))
 
     def __str__(self):
         return self.msg
 
 def print_debug(*args):
-    statement = ' '.join(args)
+    statement = ' '.join(map(str, args))
     print >>sys.stderr, statement
 
 def assert_path(path):
     '''if path does not exist, raise IOError'''
     if path is None: raise IOError('NoneType is not a valid path')
     if os.path.exists(path): return True
-    else: raise IOError(' '.join(['File or directory', path, 'not found']))
+    else:
+        raise IOError(errno.ENOENT, os.strerror(errno.ENOENT), path)
 
 class FilenameParser(object):
     @exit_on_Usage
@@ -632,7 +649,6 @@ class FilenameParser(object):
                 self.output_dir = self.target_dir
        
         if debug: print_debug('Using', self.output_dir, 'as output_dir')
-        self.check_output_dir(self.output_dir)
 
         self.protoname = os.path.splitext(
                             os.path.basename(self.input_file))[0]
@@ -710,12 +726,12 @@ def leaves(dir_or_file, allow_symlinks = True, ignore_hidden_files = True,
 
 def valid_directories(directory):
     '''wrapper for glob.glob, enforces that output must be a valid directory'''
-    directories = [defir for dir in glob.glob(directory) if os.path.isdir(dir)]
+    directories = [dir for dir in glob.glob(directory) if os.path.isdir(dir)]
     directories.reverse #to use the newest version, in case we have foo-version
     return directories
 
 @exit_on_Usage
-def path_to_executable(name, directories=None):
+def path_to_executable(name, directories=None, max_depth=2):
     """
     construct the path to the executable, search in order
     
@@ -728,28 +744,39 @@ def path_to_executable(name, directories=None):
     *we reverse the order so that we will usually get the newest version
     """
     # if name is a list, iterate over it to find exe and catch errors
+    # bug workaround
     if type(name) is list:
         for try_name in name:
-            try: path_to = path_to_executable(try_name)
-            except Usage: continue
+            try:
+                path_to = _path_to_executable(try_name,
+                                              directories=directories,
+                                              max_depth=max_depth)
+            except StandardError: continue
             return path_to
-    if type(name) is list:
         raise Usage("Could not find an executable with any of these names:",
                     ", ".join(name))
-
-    if type(directories) is list:
-       for d in directories:
-           try: path_to = path_to_executable(name, directories)
-           except Usage: continue 
-           return path_to
+    else:
+        try: path_to = _path_to_executable(name,
+                                           directories=directories,
+                                           max_depth=max_depth)
+        except StandardError:
+            raise Usage("Could not find executable " + name)
+        return path_to
+        
+def _path_to_executable(name, directories=None, max_depth=2):
+    using_windows = platform.system() == 'Windows'
 
     #try specified directory
     if directories is not None:
-        for directory in valid_directories(directories):
-            full_path = os.path.join(directory, name)
-            if os.path.exists(full_path):
-                if objects.access(full_path, os.X_OK):
+        if type(directories) is not list: directories = [directories]
+        for d in directories:
+            for directory in valid_directories(d):
+                full_path = os.path.join(directory, name)
+                if is_valid_executable(full_path):
                     return full_path
+                if using_windows and is_valid_executable(full_path + '.exe'):
+                    return full_path + '.exe'
+            
     #try PATH
     try: PATH = os.environ['PATH']
     except NameError:
@@ -757,25 +784,48 @@ def path_to_executable(name, directories=None):
         except NameError: raise Usage("Could not determine PATH")
     for p in PATH.split(os.pathsep):
         full_path = os.path.join(p, name)
-        if os.path.exists(full_path):
-            if os.access(full_path, os.X_OK):
-                return full_path
+        if is_valid_executable(full_path):
+            return full_path
+        if using_windows and is_valid_executable(full_path + '.exe'):
+            return full_path + '.exe'
             
+    #try python scripts
+    try:
+        script_path = sysconfig.get_path('scripts')
+        full_path = os.path.join(script, name)
+        if is_valid_executable(full_path):
+            return full_path
+        if using_windows and is_valid_executable(full_path + '.exe'):
+            return full_path + '.exe'
+    except NameError, AttributeError: pass
+        
     # check if we're on Windows, and try a little harder
-    if platform.system() == 'Windows':
+    if using_windows:
         all_exes = itertools.ifilter(lambda f: f.endswith('exe'),
                     itertools.chain(
-                        leaves(os.environ['PROGRAMFILES'], max_depth=2),
-                        leaves(os.environ['PROGRAMFILES(X86)'], max_depth=2)
+                                    leaves(os.environ['PROGRAMFILES'],
+                                           max_depth=max_depth),
+                                    leaves(os.environ['PROGRAMFILES(X86)'],
+                                           max_depth=max_depth)
                     ))
-        namex = name + os.extsep + 'exe'
+        namex = name + '.exe'
         for exe in all_exes:
             exename = os.path.split(exe)[1]
-            if (exename == name or exename == namex) and os.access(exe, os.X_OK):
-                return exe # success
-        
-    #give up
-    raise Usage("Could not find executable ", name)
+            if (exename == name or exename == namex) and \
+                is_valid_executable(exe):
+                    return exe # success
+    
+    # give up
+    raise StandardError
+
+def is_valid_executable(filename):
+    """
+    checks if a filename is a valid executable
+    """
+    if os.path.exists(filename):
+        if os.access(filename, os.X_OK):
+            return filename
+    return False
 
 def usage_info():
     return ' '.join(['Usage:', PROGRAM_NAME, '[OPTIONS]', 'FILE(S)'])
